/* ----------------------------------------------------------------------------
 * Raima Database Manager
 *
 * Copyright (c) 2012 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 * --------------------------------------------------------------------------*/
#include <stdio.h>

/* rdm.h is required in all files using the RDM Database */
#include "rdm.h"

/* perf01_db.h was generated from the DDL file perf01_db.sdl. */
#include "perf01_db.h"

/* perf01_db_dbd.h was generated from the DDL file perf01_db.sdl. using rdmsql with the -c option */
#include "perf01_db_dbd.h"

/*! \file perf01.c
 * \brief Source code for the RDM Perf01 example
 */
 
/*!
 * \mainpage Perf01
 * \author Jeffrey R. Parsons, Raima Sr. Engineer
 * \date 2012
 *
 * \section Description
 * EXAMPLE - Perf01
 *
 * The Perf01 example performs a number of atomic database operations
 * on a very simple database schema using the RDM Core C APIs.
 *
 * Each time you run this example
 *   * The database is initialized (removing all existing data)
 *   * 50,000 new records are created
 *   * Each record is read
 *   * Each record is updated
 *   * Each record is read again
 *   * Each record is deleted
 *   * Database is cleaned up
 *
 * \section Perf01 Database
 * The DDL (Database Definition Language) specification for the database used
 * in this example is located in the file perf01_db.sdl. The example utilizes 
 * the SQL DDL syntax and specifies an inmemory database with a single 
 * table (simple).  The table has a single column (int_col).  There are no 
 * index definitions for this database.
 *
 * \include perf01_db.sdl
 *
 * The schema was compiled using the RDM rdmsql utility with the -c option to
 * generate an embedded database dictionary (used by the RDM d_open_ptr API).
 *
 * \code rdmsql -c perf01_db.sdl \endcode
 *
 * The database dictionary information is embedded inside the application so 
 * an external DBD file stored on the file system is not required.
 *
 * \section RDM Functions Used
 * The following RDM functions are illustrated:
 *   * d_tfsinitEx/d_tfsterm
 *   * d_opentask/d_closetask
 *   * d_open_ptr/d_close
 *   * d_fillnew
 *   * d_recfrst/d_recnext
 *   * d_recread
 *   * d_recwrite
 *   * d_reclast/d_recprev
 *   * d_delete
 *   * d_errorname
 *   * d_errorstr
 *
 * They are fully documented in http://docs.raima.com/rdm/.
 */

/*! \brief Display a message and the elapsed time to perform database operations
 *
 */
void display_elapsed_time(
    const char *msg,         /**< [in] Message to be display along with the elapsed time */
    uint64_t    start_time,  /**< [in] Start of elapsed time to display */
    uint64_t    end_time)    /**< [in] End of elapsed */
{
    uint64_t elapsed_time = end_time - start_time;
    printf("%s" UINT64_SPEC " millisecs\n", msg, elapsed_time/1000000);
}

/*! \brief Initialize the RDM runtime library for use in the perf01 example
 *
 * This function initializes the RDM Transactional File Server (TFS) to use 
 * the single-user non-transactional STANDALONE implementation.  It also creates
 * an RDM runtime task and opens the "perf01" database in exclusive ('x') mode.
 * Exclusve mode does not require database locks or transactions.
 *
 * @return Returns a 32-bit RDM return code (S_OKAY if successful)
 */
int32_t initialize(
    TFS_HANDLE  *hTFS,   /**< [out] Pointer to the TFS handle to be initialized */
    DB_TASK    **task)   /**< [out] Pointer to the RDM task to be initialized */
{
    int32_t  rdm_status = S_OKAY;
    uint64_t start_time;
    uint64_t end_time;    

    /* Initialize the TFS to use the STANDALONE configuration (TFSS) */    
    start_time = psp_timeNanoSecs();
    rdm_status = d_tfsinitEx(NULL, NULL, hTFS, TFS_TYPE_STANDALONE);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error initializing the TFS (%s - %s)\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        return rdm_status;
    }

    /* Create the RDM DB_TASK */
    rdm_status = d_opentask(task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error opening the RDM task (%s - %s)\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        return rdm_status;
    }

    /* Open the perf01 database in exclusive mode (does not require locks or tranactions) */
    rdm_status = d_open_ptr("perf01_db", "x", &perf01_db_dbd, sizeof(perf01_db_dbd), *task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error opening the perf01 database (%s - %s)\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        return rdm_status;
    }

    /* Initialize the database (creates new database files truncating any existing data) */    
    rdm_status = d_initialize(*task, CURR_DB);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error initializing the perf01 database (%s - %s)\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        return rdm_status;
    }
    end_time = psp_timeNanoSecs();  
    
    /* Display information */
    printf("Initializing with STANDALONE configuration\n");
    display_elapsed_time("Preparing a new database:\t\t\t", start_time, end_time);
    printf("\tOne, simple record type.\n");
    printf("\tIn-memory\n");
    printf("\tOpened exclusively\n");

    return rdm_status;
}

/*! \brief Cleanup the RDM runtime library
 *
 * This functions closes all open databases and cleans up the RDM runtime task
 * used in the perf01 example.  It also terminates the STANDALONE TFS used
 * in the example. 
 *
 * @return Returns a 32-bit RDM return code (S_OKAY if successful)
 */
void cleanup(
    TFS_HANDLE  hTFS,   /**< [out] Pointer to the TFS handle to be terminated */
    DB_TASK    *task)   /**< [out] Pointer to the RDM task to be cleaned up */
{
    uint64_t start_time;
    uint64_t end_time;

    start_time = psp_timeNanoSecs();
    if(task)
    {
        /* Close any open databases */
        d_close(task);

        /* Close the task */
        d_closetask(task);
    }

    /* Terminate the STANDALONE TFS */
    d_tfsterm(hTFS);

    end_time = psp_timeNanoSecs();

    display_elapsed_time("Cleanup:\t\t\t\t\t", start_time, end_time);
}

/*! \brief Insert 50,000 simple records into the perf01 database
 *
 * This function adds 50,000 simple records into the inmemory perf database.  The
 * simple record contains only a single integer field with no index.  The function will
 * calculate and display the total time elapsed while inserting the records
 *
 * @return Returns a 32-bit RDM return code (S_OKAY if successful)
 */
int32_t add_records(
    DB_TASK *task)         /**< [in] RDM task initialized with the perf01 database open and as the current database */
{
    int32_t       rdm_status = S_OKAY;
    int32_t       ii;
    struct simple simple_rec;
    uint64_t      start_time;
    uint64_t      end_time;

    memset(&simple_rec, 0, sizeof(simple_rec));

    start_time = psp_timeNanoSecs();
    /* Loop to add 50,000 new simple record */
    for(ii=0; ii<50000; ii++)
    {
        /* Set values for the new record */
        simple_rec.int_col = 1;

        /* Add the record to the database */
        rdm_status = d_fillnew(SIMPLE, &simple_rec, task, CURR_DB);
        if(rdm_status != S_OKAY)
        {
            fprintf(stderr, "Error adding a simple record to the perf01 database (%s - %s)\n", d_errorname(rdm_status), d_errorstr(rdm_status));
            return rdm_status;
        }
    }
    end_time = psp_timeNanoSecs();

    display_elapsed_time("Add 50,000 records:\t\t\t\t", start_time, end_time);
    return rdm_status;
}

/*! \brief Read all simple records and calculate the sum of the int_col column
 *
 * This function iterates through and reads all simple records stored in the perf01 database.  The
 * iteration is done in raw file order through the d_recfrst/d_recnext APIs.  The function will
 * calculate the sum of the int_col field from all simple records in the database.  The function will also
 * calculate and display the total time elapsed while reading the records.
 *
 * @return Returns a 32-bit RDM return code (S_OKAY if successful)
 */
int32_t read_records(
    DB_TASK *task)         /**< [in] RDM task initialized with the perf01 database open and as the current database */
{
    int32_t       rdm_status = S_OKAY;
    uint64_t      total = 0;
    struct simple simple_rec;
    uint64_t      start_time;
    uint64_t      end_time;

    memset(&simple_rec, 0, sizeof(simple_rec));

    start_time = psp_timeNanoSecs();    
    /* Loop through each simple record in file order */
    for (rdm_status = d_recfrst (SIMPLE, task, CURR_DB);
         rdm_status == S_OKAY;
         rdm_status = d_recnext (task, CURR_DB))
    {
        /* Read the record */
        rdm_status = d_recread (&simple_rec, task, CURR_DB);
        if(rdm_status != S_OKAY)
        {
            fprintf(stderr, "Error reading a simple record from the perf01 database (%s - %s)\n", d_errorname(rdm_status), d_errorstr(rdm_status));
            return rdm_status;
        }

        /* Add the value read to the running sum */
        total += simple_rec.int_col;
    }    
    end_time = psp_timeNanoSecs();

    /* We expect rdm_status to be S_NOTFOUND when we break out of the loop */
    if(rdm_status == S_NOTFOUND)
    {
        rdm_status = S_OKAY;
    }

    
    display_elapsed_time("Scanning, reading, summing 50,000 records:\t", start_time, end_time);
    printf("\tSum = " UINT64_SPEC "\n", total);
    return rdm_status;
}

/*! \brief Update all simple records in the database
 *
 * This function iterates through and updates all simple records stored in the perf01 database.  The
 * iteration is done in raw file order through the d_recfrst/d_recnext APIs.  The function will
 * read, increment, and then update the value stored in the inc_col field.
 *
 * @return Returns a 32-bit RDM return code (S_OKAY if successful)
 */
int32_t update_records(
    DB_TASK *task)         /**< [in] RDM task initialized with the perf01 database open and as the current database */
{
    int32_t       rdm_status = S_OKAY;
    struct simple simple_rec;
    uint64_t      start_time;
    uint64_t      end_time;

    memset(&simple_rec, 0, sizeof(simple_rec));

    start_time = psp_timeNanoSecs();
    /* Loop through each simple record in file order */
    for (rdm_status = d_recfrst (SIMPLE, task, CURR_DB);
         rdm_status == S_OKAY;
         rdm_status = d_recnext (task, CURR_DB))
    {
        /* Read the entire simple record */
        rdm_status = d_recread (&simple_rec, task, CURR_DB);
        if(rdm_status != S_OKAY)
        {
            fprintf(stderr, "Error reading a simple record from the perf01 database (%s - %s)\n", d_errorname(rdm_status), d_errorstr(rdm_status));
            return rdm_status;
        }

        /* increment the inc_col field */
        simple_rec.int_col++;
        
        /* Update the simple record with the new int_col value */
        rdm_status = d_recwrite (&simple_rec, task, CURR_DB);
        if(rdm_status != S_OKAY)
        {
            fprintf(stderr, "Error updating a simple record from the perf01 database (%s - %s)\n", d_errorname(rdm_status), d_errorstr(rdm_status));
            return rdm_status;
        }
    }    
    end_time = psp_timeNanoSecs();

    /* We expect rdm_status to be S_NOTFOUND when we break out of the loop */
    if(rdm_status == S_NOTFOUND)
    {
        rdm_status = S_OKAY;
    }

    display_elapsed_time("Updating 50,000 records:\t\t\t", start_time, end_time);
    return rdm_status;
}

/*! \brief Delete all simple records from the database
 *
 * This function will iterate and delete all simple records stored in the perf01 database.  The
 * iteration is done in raw file order through the d_recfrst/d_recnext APIs.  The function will 
 * calculate and display the total time elapsed while deleting the records.
 *
 * @return Returns a 32-bit RDM return code (S_OKAY if successful)
 */
int32_t delete_records(
    DB_TASK *task)         /**< [in] RDM task initialized with the perf01 database open and as the current database */
{
    int32_t       rdm_status = S_OKAY;
    uint64_t      start_time;
    uint64_t      end_time;

    start_time = psp_timeNanoSecs();
    /* Loop through each simple record in file order */
    for (rdm_status = d_reclast (SIMPLE, task, CURR_DB);
         rdm_status == S_OKAY;
         rdm_status = d_recprev (task, CURR_DB))
    {
        /* Delete the current record */
        rdm_status = d_delete (task, CURR_DB);
        if(rdm_status != S_OKAY)
        {
            fprintf(stderr, "Error deleting a simple record from the perf01 database (%s - %s)\n", d_errorname(rdm_status), d_errorstr(rdm_status));
            return rdm_status;
        }
    }    
    end_time = psp_timeNanoSecs();

    /* We expect rdm_status to be S_NOTFOUND when we break out of the loop */
    if(rdm_status == S_NOTFOUND)
    {
        rdm_status = S_OKAY;
    }

    display_elapsed_time("Deleting 50,000 records:\t\t\t", start_time, end_time);
    return rdm_status;
}

/*! \brief Main function for perf01 example
 *
 * The function initializes the RDM environment and runs the create, read, 
 * update, and delete operations.
 *
 * @return Returns 0 on success, and 1 if there were any errors
 */
int32_t main(void)
{
    DB_TASK    *task;
    TFS_HANDLE  hTFS;
    int32_t     rdm_status;

    /* Display the example title */
    printf("%s", RDM_EXAMPLE_DESC("Perf01"));

    /* Initialize the TFS, task, and open/initialize the database */
    rdm_status = initialize(&hTFS, &task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Application is aborting due to initialization error\n");
        cleanup(hTFS, task);
        return 1;
    }

    /* Add 50,000 records to the database */
    rdm_status = add_records(task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Application is aborting due to error adding records\n");
        cleanup(hTFS, task);
        return 1;
    }

    /* Read all of the records in the database */
    rdm_status = read_records(task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Application is aborting due to error reading records\n");
        cleanup(hTFS, task);
        return 1;
    }

    /* Update all of the records in the database */
    rdm_status = update_records(task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Application is aborting due to error updating records\n");
        cleanup(hTFS, task);
        return 1;
    }

    /* Read all of the records in the database */
    rdm_status = read_records(task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Application is aborting due to error reading records\n");
        cleanup(hTFS, task);
        return 1;
    }

    /* Remove all of the records in the database */
    rdm_status = delete_records(task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Application is aborting due to error deleting records\n");
        cleanup(hTFS, task);
        return 1;
    }

    cleanup(hTFS, task);

    printf("\nPress Enter to terminate");
    getchar();

	return 0;
}
